@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2002 Zope Corporation and Contributors.
+# Copyright (c) 2002-2006 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -12,13 +12,18 @@
 #
 ##############################################################################
 """
-$Id: QueueCatalog.py,v 1.18 2004/11/19 10:34:10 ctheune Exp $
+$Id: QueueCatalog.py 68067 2006-05-09 15:31:37Z jens $
 """
 # Python std. lib
+import logging
+import sets
+import sys
 from time import time
 from types import StringType
 
 # Other packages
+from ZODB.POSException import ConflictError
+from ZEO.Exceptions import ClientDisconnected
 from zExceptions import Unauthorized
 from ExtensionClass import Base
 from OFS.SimpleItem import SimpleItem
@@ -35,7 +40,9 @@ from Acquisition import Implicit, aq_base, aq_inner, aq_parent
 # Local
 from CatalogEventQueue import CatalogEventQueue, EVENT_TYPES, ADDED_EVENTS
 from CatalogEventQueue import ADDED, CHANGED, CHANGED_ADDED, REMOVED
+from CatalogEventQueue import SAFE_POLICY, ALTERNATIVE_POLICY
 
+logger = logging.getLogger('event.QueueCatalog')
 
 _zcatalog_methods = {
     'catalog_object': 1,
@@ -51,6 +58,11 @@ _zcatalog_methods = {
     '__call__': 1,
     'refreshCatalog': 1,
     'Indexes': 1,
+    'unrestrictedSearchResults': 1,
+    'manage_addIndex': 1,
+    'manage_addColumn': 1,
+    'manage_catalogClear': 1,
+    'getIndexObjects': 1,
     }
 
 _is_zcatalog_method = _zcatalog_methods.has_key
@@ -93,18 +105,29 @@ class QueueCatalog(Implicit, SimpleItem):
     _immediate_indexes = ()  # The names of indexes to update immediately
     _location = None
     _immediate_removal = 1   # Flag: don't queue removal
+    _immediate_metadata_update = 1 # Flag: don't queue metadata creation
+    _process_all_indexes = 0 # Flag: queue-process all, not just non-immediate,
+                             #       indexes
     title = ''
 
+
     # When set, _v_catalog_cache is a tuple containing the wrapped ZCatalog
     # and the REQUEST it is bound to.
     _v_catalog_cache = None
 
-    def __init__(self, buckets=1009):
+    # As an alternative to the original queue conflict handling there is now
+    # a policy which will reduce conflicts, but at the cost of possibly having
+    # situations where items get cataloged unnecessarily. YMMV.
+    _conflict_policy = SAFE_POLICY
+
+    def __init__(self, buckets=1009, conflict_policy=SAFE_POLICY):
         self._buckets = buckets
+        self._conflict_policy = conflict_policy
         self._clearQueues()
 
     def _clearQueues(self):
-        self._queues = [CatalogEventQueue() for i in range(self._buckets)]
+        self._queues = [ CatalogEventQueue(self.getConflictPolicy()) 
+                              for i in range(self._buckets) ]
 
     def getTitle(self):
         return self.title
@@ -141,15 +164,31 @@ class QueueCatalog(Implicit, SimpleItem):
     def setImmediateIndexes(self, indexes):
         self._immediate_indexes = tuple(map(str, indexes))
 
-
     security.declareProtected(view_management_screens, 'getImmediateRemoval')
     def getImmediateRemoval(self):
         return self._immediate_removal
 
     security.declareProtected(view_management_screens, 'setImmediateRemoval')
     def setImmediateRemoval(self, flag):
-        self._immediate_removal = not not flag
+        self._immediate_removal = bool(flag)
 
+    security.declareProtected(view_management_screens,
+                              'getImmediateMetadataUpdate')
+    def getImmediateMetadataUpdate(self):
+        return self._immediate_metadata_update
+
+    security.declareProtected(view_management_screens,
+                              'setImmediateMetadataUpdate')
+    def setImmediateMetadataUpdate(self, flag):
+        self._immediate_metadata_update = bool(flag)
+
+    security.declareProtected(view_management_screens, 'getProcessAllIndexes')
+    def getProcessAllIndexes(self):
+        return self._process_all_indexes
+
+    security.declareProtected(view_management_screens, 'setProcessAllIndexes')
+    def setProcessAllIndexes(self, flag):
+        self._process_all_indexes = bool(flag)
 
     security.declareProtected(view_management_screens, 'getBucketCount')
     def getBucketCount(self):
@@ -162,6 +201,25 @@ class QueueCatalog(Implicit, SimpleItem):
         self._buckets = int(count)
         self._clearQueues()
 
+    security.declareProtected(view_management_screens, 'getConflictPolicy')
+    def getConflictPolicy(self):
+        """ Return the currently-used conflict policy
+        """
+        return self._conflict_policy
+
+    security.declareProtected(view_management_screens, 'setConflictPolicy')
+    def setConflictPolicy(self, policy=SAFE_POLICY):
+        """ Set the conflic policy to be used
+        """
+        try:
+            policy = int(policy)
+        except ValueError:
+            return
+
+        if ( policy in (SAFE_POLICY, ALTERNATIVE_POLICY) and
+             policy != self.getConflictPolicy() ):
+            self._conflict_policy = policy
+            self._clearQueues()
 
     security.declareProtected(manage_zcatalog_entries, 'getZCatalog')
     def getZCatalog(self, method=''):
@@ -189,7 +247,7 @@ class QueueCatalog(Implicit, SimpleItem):
                 raise QueueConfigurationError(
                     "ZCatalog not found at %s." % self._location
                     )
-            if not hasattr(ZC, 'getIndexObjects'):  # XXX need a better check
+            if not hasattr(ZC, 'getIndexObjects'): # XXX need a better check
                 raise QueueConfigurationError(
                     "The object at %s does not implement the "
                     "IZCatalog interface." % self._location
@@ -226,9 +284,18 @@ class QueueCatalog(Implicit, SimpleItem):
         self._queues[hash(uid) % self._buckets].update(uid, etype)
 
     security.declareProtected(manage_zcatalog_entries, 'catalog_object')
-    def catalog_object(self, obj, uid=None):
+    def catalog_object(self, obj, uid=None, idxs=None, update_metadata=1):
+        # update_metadata=0 is ignored if the queued catalog is set to
+        # update metadata during queue processing, rather than immediately
 
-        # Make sure the current context is allowed to to this:
+        # similarly, limiting the idxs only limits the immediate indexes.  If 
+        # any work needs to be done in the queue processing, it will all be
+        # done: we have not implemented partial indexing during queue 
+        # processing.  The only way to avoid any of it is to avoid all of it 
+        # (i.e., update metadata immediately and don't have any indexes to 
+        # update on the queued side).
+
+        # Make sure the current context is allowed to do this:
         catalog_object = self.getZCatalog('catalog_object')
 
         if uid is None:
@@ -237,10 +304,9 @@ class QueueCatalog(Implicit, SimpleItem):
             uid = '/'.join(uid)
 
         catalog = self.getZCatalog()
-        cat_indexes = list(catalog.indexes())
-        cat_indexes.sort()
-        immediate_indexes = list(self._immediate_indexes)
-        immediate_indexes.sort()
+        cat_indexes = sets.Set(catalog.indexes())
+        immediate_indexes = sets.Set(self._immediate_indexes)
+        cat_indexes -= immediate_indexes
 
         # The ZCatalog API doesn't allow us to distinguish between
         # adds and updates, so we have to try to figure this out
@@ -256,25 +322,33 @@ class QueueCatalog(Implicit, SimpleItem):
         # down a bit, but adds should be relatively infrequent.
 
         # Now, try to decide if the catalog has the uid (path).
+        already_cataloged = cataloged(catalog, uid)
+        if not already_cataloged:
+            # Looks like we should add, but maybe there's already a
+            # pending add event. We'd better check the event queue:
+            already_cataloged = (
+                self._queues[hash(uid) % self._buckets].getEvent(uid) in
+                ADDED_EVENTS)
 
-        if immediate_indexes != cat_indexes:
-            if cataloged(catalog, uid):
-                event = CHANGED
-            else:
-                # Looks like we should add, but maybe there's already a
-                # pending add event. We'd better check the event queue:
-                if (self._queues[hash(uid) % self._buckets].getEvent(uid) in
-                    ADDED_EVENTS):
-                    event = CHANGED
-                else:
-                    event = ADDED
+        if idxs and already_cataloged:
+            # if not already_cataloged, we index the whole thing
+            idxs = sets.Set(idxs)
+            immediate_indexes.intersection_update(idxs)
+            cat_indexes.intersection_update(idxs)
 
-            self._update(uid, event)
+        immediate_metadata = self.getImmediateMetadataUpdate()
+        if cat_indexes or update_metadata and not immediate_metadata:
+            self._update(uid, already_cataloged and CHANGED or ADDED)
 
         if immediate_indexes:
             # Update some of the indexes immediately.
-            catalog.catalog_object(obj, uid, immediate_indexes)
-
+            catalog.catalog_object(
+                obj, uid, immediate_indexes,
+                update_metadata=update_metadata and immediate_metadata)
+        elif update_metadata and immediate_metadata:
+            # if it is added, no point in doing the metadata, and it will be
+            # done in the queue process anyway
+            catalog._catalog.updateMetadata(obj, uid)
 
     security.declareProtected(manage_zcatalog_entries, 'uncatalog_object')
     def uncatalog_object(self, uid):
@@ -284,7 +358,9 @@ class QueueCatalog(Implicit, SimpleItem):
         self._update(uid, REMOVED)
 
         if self._immediate_removal:
-            self._process_queue(self._queues[hash(uid) % self._buckets], limit=None)
+            self._process_queue( self._queues[hash(uid) % self._buckets]
+                               , limit=None
+                               )
 
     security.declareProtected(manage_zcatalog_entries, 'process')
     def process(self, max=None):
@@ -312,13 +388,28 @@ class QueueCatalog(Implicit, SimpleItem):
     def _process_queue(self, queue, limit):
         """Process a single queue"""
         catalog = self.getZCatalog()
+
+        if self.getProcessAllIndexes():
+            idxs = None
+        else:
+            cat_indexes = sets.Set(catalog.indexes())
+            immediate_indexes = sets.Set(self._immediate_indexes)
+            if not immediate_indexes or immediate_indexes==cat_indexes:
+                idxs = None # do all of 'em
+            else:
+                idxs = list(cat_indexes - immediate_indexes)
         events = queue.process(limit)
         count = 0
 
         for uid, (t, event) in events.items():
             if event is REMOVED:
-                if cataloged(catalog, uid):
-                    catalog.uncatalog_object(uid)
+                try:
+                    if cataloged(catalog, uid):
+                        catalog.uncatalog_object(uid)
+                except (ConflictError, ClientDisconnected):
+                    raise
+                except:
+                    logger.error('error uncataloging object', exc_info=True)
             else:
                 # add or change
                 if event is CHANGED and not cataloged(catalog, uid):
@@ -326,7 +417,15 @@ class QueueCatalog(Implicit, SimpleItem):
                 # Note that the uid may be relative to the catalog.
                 obj = catalog.unrestrictedTraverse(uid, None)
                 if obj is not None:
-                    catalog.catalog_object(obj, uid)
+                    immediate_metadata = self.getImmediateMetadataUpdate()
+                    try:
+                        catalog.catalog_object(
+                            obj, uid, idxs=idxs,
+                            update_metadata=not immediate_metadata)
+                    except (ConflictError, ClientDisconnected):
+                        raise
+                    except:
+                        logger.error('error cataloging object', exc_info=True)
 
             count = count + 1
 
@@ -349,26 +448,23 @@ class QueueCatalog(Implicit, SimpleItem):
         self.uncatalog_object(self.uidForObject(object))
 
     security.declarePrivate('reindexObject')
-    def reindexObject(self, object, idxs=[]):
+    def reindexObject(self, object, idxs=None,update_metadata=1,uid=None):
         """Update catalog after object data has changed.
 
         The optional idxs argument is a list of specific indexes
         to update (all of them by default).
         """
-        # Punt for now and ignore idxs.
-        self.catalog_object(object, self.uidForObject(object))
+        self.catalog_object(object, uid or self.uidForObject(object), idxs=idxs,
+                            update_metadata=update_metadata)
 
     security.declarePrivate('uidForObject')
     def uidForObject(self, obj):
         """Get a catalog uid for the object. Allows the underlying catalog
         to determine the uids if it implements this method"""
         catalog = self.getZCatalog()
-        try:
-            uidForObject = aq_base(catalog).uidForObject
-        except AttributeError:
-            return '/'.join(obj.getPhysicalPath())
-        else:
-            return uidForObject(obj)
+        if hasattr(aq_base(catalog), 'uidForObject'):
+            return catalog.uidForObject(obj)
+        return '/'.join(obj.getPhysicalPath())
 
     # Provide web pages. It would be nice to use views, but Zope 2.6
     # just isn't ready for views. :( In particular, we'd have to fake
@@ -384,12 +480,16 @@ class QueueCatalog(Implicit, SimpleItem):
 
     security.declareProtected(view_management_screens, 'manage_edit')
     def manage_edit(self, title='', location='', immediate_indexes=(),
-                    immediate_removal=0, bucket_count=0, RESPONSE=None):
+                    immediate_removal=0, bucket_count=0, immediate_metadata=0,
+                    all_indexes=0, conflict_policy=SAFE_POLICY, RESPONSE=None):
         """ Edit the instance """
         self.title = title
         self.setLocation(location or None)
         self.setImmediateIndexes(immediate_indexes)
         self.setImmediateRemoval(immediate_removal)
+        self.setImmediateMetadataUpdate(immediate_metadata)
+        self.setProcessAllIndexes(all_indexes)
+        self.setConflictPolicy(conflict_policy)
         if bucket_count:
             bucket_count = int(bucket_count)
             if bucket_count != self.getBucketCount():
@@ -462,8 +562,7 @@ class QueueCatalog(Implicit, SimpleItem):
     security.declarePublic('getTitle', 'title_or_id')
 
     security.declareProtected(manage_zcatalog_entries,
-                              'catalog_object', 'uncatalog_object',
-                              'refreshCatalog')
+                              'catalog_object', 'uncatalog_object')
 
 
 def cataloged(catalog, path):
